==> python-plugin-3.9.4-sources.jar: com/xebialabs/deployit/plugin/python/PythonStep.java <==
package com.xebialabs.deployit.plugin.python;

import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import com.google.common.io.Closeables;

import com.xebialabs.deployit.plugin.api.flow.ExecutionContext;
import com.xebialabs.deployit.plugin.api.flow.Preview;
import com.xebialabs.deployit.plugin.api.flow.PreviewStep;
import com.xebialabs.deployit.plugin.api.flow.StepExitCode;
import com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler;
import com.xebialabs.overthere.OverthereConnection;
import com.xebialabs.overthere.OverthereFile;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.io.Closeables.closeQuietly;
import static com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler.handleStderr;
import static com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler.handleStdout;
import static com.xebialabs.deployit.plugin.python.PythonManagingContainer.CONNECT_FROM_STAND_ALONE_SCRIPT;
import static com.xebialabs.deployit.plugin.python.PythonManagingContainer.DISCONNECT_FROM_STAND_ALONE_SCRIPT;
import static com.xebialabs.deployit.plugin.remoting.preview.PreviewOverthereConnection.getPreviewConnection;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.MDC_KEY_SCRIPT_PATH;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.appendScriptDir;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.appendScripts;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.dumpScript;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.loadScript;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.uploadScript;
import static com.xebialabs.overthere.util.OverthereUtils.getName;

@SuppressWarnings("serial")
abstract class PythonStep implements PreviewStep {

    private static final String BASE_RUNTIME_PATH = "python/runtime";

    private final PythonManagingContainer container;
    private final String scriptPath;
    private final Map<String, Object> pythonVars;
    private final String description;
    private boolean uploadArtifactData = true;
    private Preview preview;

    public PythonStep(PythonManagingContainer container, String scriptPath, Map<String, Object> pythonVars, String description) {
        this.container = checkNotNull(container, "container is null");
        this.scriptPath = checkNotNull(scriptPath, "scriptPath is null");
        this.pythonVars = checkNotNull(pythonVars, "pythonVars is null");
        this.description = checkNotNull(description, "description is null");
    }

    protected StepExitCode doExecute(ExecutionContext ctx) throws Exception {
        MDC.put(MDC_KEY_SCRIPT_PATH, scriptPath);
        try {
            OverthereConnection conn = container.getHost().getConnection();
            try {
                String generatedScript = aggregateScript(conn);
                dumpPythonScript(getName(scriptPath), generatedScript);
                OverthereFile uploadedScriptFile = uploadScript(conn, scriptPath, generatedScript);

                int res = executePythonScript(ctx, conn, uploadedScriptFile);

                logger.debug("Exit code: {}", res);
                if(res == 0) {
                    return StepExitCode.SUCCESS;
                } else {
                    return StepExitCode.FAIL;
                }
            } finally {
                Closeables.closeQuietly(conn);
            }
        } finally {
            MDC.remove(MDC_KEY_SCRIPT_PATH);
        }
    }

    @Override
    public Preview getPreview() {
        if(preview == null) {
            OverthereConnection poc = getPreviewConnection();
            try {
                String contents = aggregateScript(poc);
                preview = Preview.withSourcePathAndContents(scriptPath, contents);
            } finally {
                poc.close();
            }
        }
        return preview;
    }

    protected String aggregateScript(OverthereConnection connection) {
        String pythonVarsPython = PythonVarsConverter.javaToPython(connection, pythonVars, uploadArtifactData);
        StringBuilder b = new StringBuilder();

        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            appendRuntimeScripts(container, b);
        }

        appendDeployedRuntimeScripts(b);

        b.append("# PythonVars\n");
        b.append(pythonVarsPython);
        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            b.append("#\n" + CONNECT_FROM_STAND_ALONE_SCRIPT + "()\n");
        }
        b.append(loadScript(scriptPath));
        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            b.append("#\n" + DISCONNECT_FROM_STAND_ALONE_SCRIPT + "()\n");
        }
        return b.toString();
    }

    static void appendRuntimeScripts(PythonManagingContainer c, StringBuilder b) {
        appendBaseRuntimeScripts(b);
        appendContainerRuntimeScripts(c, b);
        appendContainerLibraryScripts(c, b);
    }

    private static void appendBaseRuntimeScripts(StringBuilder b) {
        appendScriptDir(BASE_RUNTIME_PATH, b);
    }

    private static void appendContainerRuntimeScripts(PythonManagingContainer c, StringBuilder b) {
        appendScriptDir(c.getRuntimePath(), b);
    }

    private static void appendContainerLibraryScripts(PythonManagingContainer c, StringBuilder b) {
        if (c.hasProperty("libraryScripts")) {
            List<String> scripts = c.getProperty("libraryScripts");
            appendScripts(scripts, b);
        }
    }

    private void appendDeployedRuntimeScripts(StringBuilder b) {
        PythonManagedDeployed<?,?> deployed = (PythonManagedDeployed<?, ?>) pythonVars.get("deployed");
        deployed = deployed == null ? (PythonManagedDeployed<?, ?>) pythonVars.get("prototype") : deployed;
        if (deployed != null) {
            for(String script: deployed.getLibraryScripts()) {
                b.append(loadScript(script));
            }
        }
    }

    private int executePythonScript(final ExecutionContext ctx, final OverthereConnection conn, final OverthereFile script) {
        if (container.runWithDaemon() && conn.canStartProcess()) {
            return getDaemon(ctx).executePythonScript(ctx, script);
        } else {
            return executePythonScriptDirectly(ctx, conn, script);
        }
    }

    protected int executePythonScriptDirectly(final ExecutionContext ctx, final OverthereConnection conn, final OverthereFile script) {
        logger.info("Executing Python script {} on {} (without daemon)", script, conn);
        DefaultExecutionOutputHandler stdoutHandler = handleStdout(ctx);
        DefaultExecutionOutputHandler stderrHandler = handleStderr(ctx);
        try {
            return conn.execute(stdoutHandler, stderrHandler, container.getScriptCommandLine(script));
        } finally {
            closeQuietly(stdoutHandler);
            closeQuietly(stderrHandler);
        }
    }

    private PythonDaemon getDaemon(final ExecutionContext context) {
        String key = "DAEMON_" + container.getId();
        PythonDaemon daemon = (PythonDaemon) context.getAttribute(key);
        if (daemon == null || !daemon.isAlive()) {
            daemon = new PythonDaemon(container);
            daemon.start(context);
            context.setAttribute(key, daemon);
        }
        return daemon;
    }

    String getScriptPath() {
        return scriptPath;
    }

    Map<String, Object> getPythonVars() {
        return pythonVars;
    }

    public void setUploadArtifactData(boolean uploadArtifactData) {
        this.uploadArtifactData = uploadArtifactData;
    }

    @Override
    public String getDescription() {
        return description;
    }

    @Override
    public String toString() {
        return scriptPath;
    }

    public static void dumpPythonScript(String scriptName, String scriptContents) {
        dumpScript(scriptName, scriptContents, scriptsLogger);
    }

    private static final Logger logger = LoggerFactory.getLogger(PythonStep.class);

    private static final Logger scriptsLogger = LoggerFactory.getLogger("com.xebialabs.deployit.plugin.python.scripts");

}

==> python-plugin-3.9.6-sources.jar: com/xebialabs/deployit/plugin/python/PythonStep.java <==
package com.xebialabs.deployit.plugin.python;

import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import com.google.common.io.Closeables;

import com.xebialabs.deployit.plugin.api.flow.ExecutionContext;
import com.xebialabs.deployit.plugin.api.flow.Preview;
import com.xebialabs.deployit.plugin.api.flow.PreviewStep;
import com.xebialabs.deployit.plugin.api.flow.StepExitCode;
import com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler;
import com.xebialabs.overthere.OverthereConnection;
import com.xebialabs.overthere.OverthereFile;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.io.Closeables.closeQuietly;
import static com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler.handleStderr;
import static com.xebialabs.deployit.plugin.overthere.DefaultExecutionOutputHandler.handleStdout;
import static com.xebialabs.deployit.plugin.python.PythonManagingContainer.CONNECT_FROM_STAND_ALONE_SCRIPT;
import static com.xebialabs.deployit.plugin.python.PythonManagingContainer.DISCONNECT_FROM_STAND_ALONE_SCRIPT;
import static com.xebialabs.deployit.plugin.remoting.preview.PreviewOverthereConnection.getPreviewConnection;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.MDC_KEY_SCRIPT_PATH;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.appendScriptDir;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.appendScripts;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.dumpScript;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.loadScript;
import static com.xebialabs.deployit.plugin.remoting.scripts.ScriptUtils.uploadScript;
import static com.xebialabs.overthere.util.OverthereUtils.getName;

@SuppressWarnings("serial")
abstract class PythonStep implements PreviewStep {

    private static final String BASE_RUNTIME_PATH = "python/runtime";

    private final PythonManagingContainer container;
    private final String scriptPath;
    private final Map<String, Object> pythonVars;
    private final String description;
    private boolean uploadArtifactData = true;
    private Preview preview;

    public PythonStep(PythonManagingContainer container, String scriptPath, Map<String, Object> pythonVars, String description) {
        this.container = checkNotNull(container, "container is null");
        this.scriptPath = checkNotNull(scriptPath, "scriptPath is null");
        this.pythonVars = checkNotNull(pythonVars, "pythonVars is null");
        this.description = checkNotNull(description, "description is null");
    }

    protected StepExitCode doExecute(ExecutionContext ctx) throws Exception {
        MDC.put(MDC_KEY_SCRIPT_PATH, scriptPath);
        try {
            OverthereConnection conn = container.getHost().getConnection();
            try {
                String generatedScript = aggregateScript(conn, false);
                dumpPythonScript(getName(scriptPath), generatedScript);
                OverthereFile uploadedScriptFile = uploadScript(conn, scriptPath, generatedScript);

                int res = executePythonScript(ctx, conn, uploadedScriptFile);

                logger.debug("Exit code: {}", res);
                if(res == 0) {
                    return StepExitCode.SUCCESS;
                } else {
                    return StepExitCode.FAIL;
                }
            } finally {
                Closeables.closeQuietly(conn);
            }
        } finally {
            MDC.remove(MDC_KEY_SCRIPT_PATH);
        }
    }

    @Override
    public Preview getPreview() {
        if(preview == null) {
            OverthereConnection poc = getPreviewConnection();
            try {
                String contents = aggregateScript(poc, true);
                preview = Preview.withSourcePathAndContents(scriptPath, contents);
            } finally {
                poc.close();
            }
        }
        return preview;
    }

    protected String aggregateScript(OverthereConnection connection, boolean maskPasswords) {
        String pythonVarsPython = PythonVarsConverter.javaToPython(connection, pythonVars, uploadArtifactData, maskPasswords);
        StringBuilder b = new StringBuilder();

        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            appendRuntimeScripts(container, b);
        }

        appendDeployedRuntimeScripts(b);

        b.append("# PythonVars\n");
        b.append(pythonVarsPython);
        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            b.append("#\n" + CONNECT_FROM_STAND_ALONE_SCRIPT + "()\n");
        }
        b.append(loadScript(scriptPath));
        if(!container.runWithDaemon() || !connection.canStartProcess()) {
            b.append("#\n" + DISCONNECT_FROM_STAND_ALONE_SCRIPT + "()\n");
        }
        return b.toString();
    }

    static void appendRuntimeScripts(PythonManagingContainer c, StringBuilder b) {
        appendBaseRuntimeScripts(b);
        appendContainerRuntimeScripts(c, b);
        appendContainerLibraryScripts(c, b);
    }

    private static void appendBaseRuntimeScripts(StringBuilder b) {
        appendScriptDir(BASE_RUNTIME_PATH, b);
    }

    private static void appendContainerRuntimeScripts(PythonManagingContainer c, StringBuilder b) {
        appendScriptDir(c.getRuntimePath(), b);
    }

    private static void appendContainerLibraryScripts(PythonManagingContainer c, StringBuilder b) {
        if (c.hasProperty("libraryScripts")) {
            List<String> scripts = c.getProperty("libraryScripts");
            appendScripts(scripts, b);
        }
    }

    private void appendDeployedRuntimeScripts(StringBuilder b) {
        PythonManagedDeployed<?,?> deployed = (PythonManagedDeployed<?, ?>) pythonVars.get("deployed");
        deployed = deployed == null ? (PythonManagedDeployed<?, ?>) pythonVars.get("prototype") : deployed;
        if (deployed != null) {
            for(String script: deployed.getLibraryScripts()) {
                b.append(loadScript(script));
            }
        }
    }

    private int executePythonScript(final ExecutionContext ctx, final OverthereConnection conn, final OverthereFile script) {
        if (container.runWithDaemon() && conn.canStartProcess()) {
            return getDaemon(ctx).executePythonScript(ctx, script);
        } else {
            return executePythonScriptDirectly(ctx, conn, script);
        }
    }

    protected int executePythonScriptDirectly(final ExecutionContext ctx, final OverthereConnection conn, final OverthereFile script) {
        logger.info("Executing Python script {} on {} (without daemon)", script, conn);
        DefaultExecutionOutputHandler stdoutHandler = handleStdout(ctx);
        DefaultExecutionOutputHandler stderrHandler = handleStderr(ctx);
        try {
            return conn.execute(stdoutHandler, stderrHandler, container.getScriptCommandLine(script));
        } finally {
            closeQuietly(stdoutHandler);
            closeQuietly(stderrHandler);
        }
    }

    private PythonDaemon getDaemon(final ExecutionContext context) {
        String key = "DAEMON_" + container.getId();
        PythonDaemon daemon = (PythonDaemon) context.getAttribute(key);
        if (daemon == null || !daemon.isAlive()) {
            daemon = new PythonDaemon(container);
            daemon.start(context);
            context.setAttribute(key, daemon);
        }
        return daemon;
    }

    String getScriptPath() {
        return scriptPath;
    }

    Map<String, Object> getPythonVars() {
        return pythonVars;
    }

    public void setUploadArtifactData(boolean uploadArtifactData) {
        this.uploadArtifactData = uploadArtifactData;
    }

    @Override
    public String getDescription() {
        return description;
    }

    @Override
    public String toString() {
        return scriptPath;
    }

    public static void dumpPythonScript(String scriptName, String scriptContents) {
        dumpScript(scriptName, scriptContents, scriptsLogger);
    }

    private static final Logger logger = LoggerFactory.getLogger(PythonStep.class);

    private static final Logger scriptsLogger = LoggerFactory.getLogger("com.xebialabs.deployit.plugin.python.scripts");

}
